/* evcnt -- Count and log keypress events
   Copyright (C) 1995 Per Cederqvist <ceder@lysator.liu.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xresource.h>

#define NITEMS(x) (sizeof(x)/sizeof((x)[0]))

XrmOptionDescRec tbl[] = {
  {"-display", ".display", XrmoptionSepArg, ""},
  {"-name",    ".name",    XrmoptionSepArg, "evcnt"},
};

static char *
get_resource(XrmDatabase db,
	     char *prog_name,
	     char *resource_name)
{
  char *name;
  char *the_class;
  int len;
  int up;
  char *s;
  XrmValue val;
  char *typ;

  len = strlen(prog_name) + strlen(resource_name) + 2;
  name = malloc(len);
  the_class = malloc(len);
  if (name == NULL || the_class == NULL)
    {
      fprintf(stderr, "Out of memory\n");
      exit(1);
    }

  sprintf(name, "%s.%s", prog_name, resource_name);
  sprintf(the_class, "%s.%s", prog_name, resource_name);

  for (s = the_class, up = 1; *s; s++)
    if (*s == '.')
      up = 1;
    else if (up) 
      {
	*s = toupper(*s);
	up = 0;
      }

  if (XrmGetResource(db, name, the_class, &typ, &val))
    {
      if (strcmp(typ, "String"))
	{
	  fprintf(stderr, "Unexpected type for name: %s\n", typ);
	  exit(1);
	}
      return strdup(val.addr);
    }
  return NULL;
}

void
recurse_win(Display *dpy,
	    Window w)
{
  Window *children;
  Window root;
  Window parent;
  int nchild;
  int ix;

  XSelectInput(dpy, w, KeyPressMask|SubstructureNotifyMask);
  if (XQueryTree(dpy, w, &root, &parent, &children, &nchild))
    {
      for (ix = 0; ix < nchild; ix++)
	recurse_win(dpy, children[ix]);

      XFree((char *) children);
    }
}

void
downcase(char *s)
{
  for (; *s; s++)
    if (isascii(*s) && isupper(*s))
      *s = tolower(*s);
}

void
do_key_event(XEvent *ev)
{
  XKeyEvent *e = (XKeyEvent*)ev;
  KeySym sym;
  char *sym_name;
  int n;
  int ix;
  unsigned char str[256+1];
  static unsigned epoch = -1;

  if (epoch == -1)
    epoch = e->time;
  
  n = XLookupString(e, str, 256, &sym, NULL);
  if (sym == NoSymbol)
    sym_name = "NoSymbol";
  else if (!(sym_name = XKeysymToString(sym)))
    sym_name = "(no name)";
  printf("%lu %u %s \"", e->time - epoch, e->keycode, sym_name);
  epoch = e->time;
  for (ix = 0; ix < n && ix < 256; ix++)
    {
      switch (str[ix])
	{
	case '"':
	  printf("\\\"");
	  break;
	case '\n':
	  printf("\\n");
	  break;
	case '\t':
	  printf("\\t");
	  break;
	case '\f':
	  printf("\\f");
	  break;
	case '\a':
	  printf("\\a");
	  break;
	case '\\':
	  printf("\\\\");
	  break;
	default:
	  if (str[ix] < 32 || str[ix] > 126 && str[ix] < 160 || str[ix] > 254)
	    printf("\\%03o", str[ix]);
	  else
	    printf("%c", str[ix]);
	  break;
	}
    }
  printf("\"\n");
}

main(int argc,
     char **argv)
{
  Display *dpy;
  XEvent ev;
  XrmDatabase db;
  XrmDatabase rmdb;
  char *display;
  char *name;

  XrmInitialize();
  db = NULL;
  XrmParseCommand(&db, tbl, NITEMS(tbl), "evcnt", &argc, argv);
  
  if (argc > 1) 
    {
      fprintf(stderr, "Bad option: %s\n", argv[1]);
      exit(1);
    }

  name = get_resource(db, "evcnt", "name");
  if (name == NULL)
    name = "evcnt";
  else
    downcase(name);
  display = get_resource(db, name, "display");
  
  dpy = XOpenDisplay(display);
  if (dpy == NULL)
    {
      fprintf(stderr, "Failed to open %s", display);
      exit(1);
    }

  recurse_win(dpy, RootWindow(dpy, DefaultScreen(dpy)));

  while(1)
    {
      XNextEvent(dpy, &ev);
      switch (ev.type) 
	{
	case KeyPress:
	  do_key_event(&ev);
	  break;
	case CreateNotify:
	  recurse_win(dpy, RootWindow(dpy, DefaultScreen(dpy)));
	  break;
	}
    }
}

/* eof */
